/****************************************************************
*																*
*		Loadable RuntimeSpaceeXtender							*
*		2.4.x porting											*
*		by IhaQueR												*
*																*
****************************************************************/




#include <rsx.h>

#define __NO_VERSION__




//	code from OWL project, www.openwall.org
static unsigned long *get_reg(struct pt_regs *regs, unsigned char regnum)
{
	switch (regnum) {
		case 0: return &regs->eax;
		case 1: return &regs->ecx;
		case 2: return &regs->edx;
		case 3: return &regs->ebx;
		case 4: return &regs->esp;
		case 5: return &regs->ebp;
		case 6: return &regs->esi;
		case 7: return &regs->edi;
	}

	return NULL;
}


static unsigned long get_modrm(struct pt_regs *regs, int *err)
{
	unsigned char modrm, sib;
	signed char rel8;
	unsigned long rel32;
	int size, regnum, scale;
	unsigned long index, base, addr, value;

	*err |= __get_user(modrm, (unsigned char *)(regs->eip + 1));
	size = 2;
	regnum = modrm & 7;
	addr = *get_reg(regs, regnum);
	if (regnum == 4 && (modrm & 0xC0) != 0xC0) {
		*err |= __get_user(sib, (unsigned char *)(regs->eip + 2));
		size = 3;
		scale = sib >> 6;
		index = *get_reg(regs, (sib >> 3) & 7);
		base = *get_reg(regs, sib & 7);
		addr = base + (index << scale);
	}

	switch (modrm & 0xC0) {
	case 0x00:
		if (regnum == 5) {
			*err |= __get_user(addr,
				(unsigned long *)(regs->eip + 2));
			size = 6;
		}
		*err |= __get_user(value, (unsigned long *)addr);
		break;

	case 0x40:
		*err |= __get_user(rel8, (signed char *)(regs->eip + size));
		size++;
		addr += rel8;
		*err |= __get_user(value, (unsigned long *)addr);
		break;

	case 0x80:
		*err |= __get_user(rel32, (unsigned long *)(regs->eip + size));
		size += 4;
		addr += rel32;
		*err |= __get_user(value, (unsigned long *)addr);
		break;

	case 0xC0:
	default:
		value = addr;
	}

	if (*err) return 0;
	regs->eip += size;
	return value;
}


int rsx_trampoline(struct pt_regs* regs)
{
int err, count;
register unsigned char insn;
unsigned long addr;


		if((regs->xcs & 0xFFFF)==RSXLDTSEG && *((unsigned char*)(regs->eip)) == 0xff) {

			if(get_user(insn, ((unsigned char*)(regs->eip)) + 1))
				return FALSE;

			if((insn & 0x38)==0x10 && insn!=0xd4) {

				err = 0;
				addr = get_modrm(regs, &err);

				if(addr > RSX_STACKLIMIT || addr < (RSX_STACKLIMIT-RSX_STACKSIZE))
					return FALSE;
#ifdef RSX_HAVEPROC
		rsx_stat_trampolines++;
#endif
				if(!err) {
					regs->esp -= 4;
					err = __put_user(regs->eip, (unsigned long *)regs->esp);
					regs->eip = addr;
				}

				count = 0;

				while(!err && !__get_user(insn, (unsigned char *)regs->eip++)) {

					if((insn & 0xF8) == 0xB8) {
						if (count++ >= 8) break;
						err |= __get_user(addr, (unsigned long *)regs->eip);
						regs->eip += 4;
						*get_reg(regs, insn & 7) = addr;
					}
					else if(insn == 0xFF) {
						err |= __get_user(insn, (unsigned char *)regs->eip);
						if((insn & 0xF8) == 0xE0) {
							regs->eip = *get_reg(regs, insn & 7);
							if(err)
								break;
							else
								return TRUE;
						}
						break;
					}
					else if(insn == 0xE9) {
						err |= __get_user(addr, (unsigned long *)regs->eip);
						if (err) break;
						regs->eip += 4 + addr;
						return TRUE;
					}
					else {
						break;
					}
				}
			}
		}

return FALSE;
}