/****************************************************************
*																*
*		Loadable RuntimeSpaceeXtender							*
*		2.4.x porting											*
*		by IhaQueR												*
*																*
****************************************************************/




#ifndef __RSX_H_INC__
#define __RSX_H_INC__



#include <linux/kernel.h>
#include <linux/module.h>



#ifndef KERNEL_VERSION
	#define KERNEL_VERSION(a,b,c) ((a)*65536+(b)*256+(c))
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	#error "THIS MODULE IS FOR THE 2.4.x KERNEL!"
#endif


#include <sys/syscall.h>

#include <linux/mm.h>
#include <linux/unistd.h>
#include <linux/spinlock.h>
#include <linux/smp_lock.h>
#include <linux/slab.h>
#include <linux/mman.h>
#include <linux/pagemap.h>
#include <linux/proc_fs.h>

#include <asm/io.h>
#include <asm/desc.h>
#include <asm/pgalloc.h>
#include <asm/desc.h>
#include <asm/uaccess.h>

#include <syslog.h>



//	set to 1 for debugging
#ifndef NODEB
	extern int MODDEBUG;
#endif

#ifdef NODEB
	#define _RDEB if(0)
	#define _STATIC static
#else
	#define _RDEB if(MODDEBUG)
	#define _STATIC static
#endif


#define FALSE 0
#define TRUE 1


//	flag for marking vm_areas as remapped
#define VM_UNMAPPED	0x40000000

//	code selector for LDT
//	index=1, type=ldt, rpl=3
#define RSXLDTSEG (0x08 | 0x04 | 0x03)



#define PUTTAB(_OCALL, _MCALL, _CALLNR) \
		_OCALL = sys_call_table[_CALLNR]; \
		sys_call_table[_CALLNR] = _MCALL;


#define RMTAB(_OCALL, _MCALL, _CALLNR) \
		if (sys_call_table[_CALLNR] != _MCALL) { \
			_RDEB printk("%s: Somebody else also played with %s\n", #_OCALL, MODNAME);\
		} \
		sys_call_table[_CALLNR] = _OCALL;


#define MLOG(MSG) _RDEB printk("%s: %s\n", MODNAME, MSG);


#define _set_tssldt_desc(n,addr,limit,type) \
__asm__ __volatile__ ("movw %w3,0(%2)\n\t" \
	"movw %%ax,2(%2)\n\t" \
	"rorl $16,%%eax\n\t" \
	"movb %%al,4(%2)\n\t" \
	"movb %4,5(%2)\n\t" \
	"movb $0,6(%2)\n\t" \
	"movb %%ah,7(%2)\n\t" \
	"rorl $16,%%eax" \
	: "=m"(*(n)) : "a" (addr), "r"(n), "ir"(limit), "i"(type))



//		our debugging name
extern const char* MODNAME;


extern void *sys_call_table[];
extern inline void avl_insert (struct vm_area_struct * new_node, struct vm_area_struct ** ptree);

typedef int (*EHOOK)(struct pt_regs*, long);

extern EHOOK rsx_gp_hook;
extern EHOOK rsx_pf_hook;
extern EHOOK rsx_log_hook;

typedef int (*SHOOK)(void*);

extern SHOOK rsx_sig_hook;
extern SHOOK rsx_rtsig_hook;


extern asmlinkage int (*original_fork)(struct pt_regs regs);
extern asmlinkage int (*original_sys_execve)(struct pt_regs regs);
extern asmlinkage int (*original_sys_mmap2)(unsigned long addr, unsigned long len,
						unsigned long prot, unsigned long flags,
						unsigned long fd, unsigned long pgoff);

						
extern struct desc_struct default_ldt[4];
extern struct desc_struct* my_gdt_table;
extern kmem_cache_t *my_vm_area_cachep;
extern unsigned long mmu_cr4_features;
extern pgprot_t protection_map[16];

extern void set_ldt_desc(unsigned int n, void *addr, unsigned int size);
extern pgprot_t protection_map[16];
extern void avl_insert_neighbours(struct vm_area_struct * new_node, struct vm_area_struct ** ptree, struct vm_area_struct ** to_the_left, struct vm_area_struct ** to_the_right);
extern void lock_vma_mappings(struct vm_area_struct *vma);
extern void unlock_vma_mappings(struct vm_area_struct *vma);
extern void build_mmap_avl(struct mm_struct * mm);
extern void __insert_vm_struct(struct mm_struct *mm, struct vm_area_struct *vmp);
extern void insert_vm_struct(struct mm_struct *mm, struct vm_area_struct *vmp);

extern int rsx_trampoline(struct pt_regs* regs);
extern void rsx_cleanup_proc(void);
extern int rsx_init_proc(void);


extern unsigned rsx_stat_exec;
extern unsigned rsx_stat_execok;
extern unsigned rsx_stat_execremap;

extern unsigned rsx_stat_fork;
extern unsigned rsx_stat_forkok;
extern unsigned rsx_stat_forkremap;

extern unsigned rsx_stat_mmap;
extern unsigned rsx_stat_mmapok;
extern unsigned rsx_stat_mmapremap;

extern unsigned rsx_stat_mmap2;
extern unsigned rsx_stat_mmap2ok;
extern unsigned rsx_stat_mmap2remap;

extern unsigned rsx_stat_gp;

extern unsigned rsx_stat_trampolines;
extern unsigned rsx_stat_trampolinesok;
extern unsigned rsx_stat_trampolinesbad;

extern unsigned rsx_stat_vmacalls;
extern unsigned rsx_stat_vmaremaps;
extern unsigned rsx_stat_vmafixes;

extern unsigned rsx_stat_kills;



struct mmap_arg_struct {
	unsigned long addr;
	unsigned long len;
	unsigned long prot;
	unsigned long flags;
	unsigned long fd;
	unsigned long offset;
};



//	Wall...
int			init_module(void);
void 		cleanup_module(void);
inline int 	rsx_is_remappable(void);
unsigned* 	rsx_get_gdt(void);
void 		rsx_load_gdt(void);
int 		rsx_map_writable_gdt(void);
int 		rsx_unmap_writable_gdt(void);
void 		rsx_flush_tlb(void);


#include <config.h>
#include <libmod.h>


#endif