/****************************************************************
*																*
*		Loadable RuntimeSpaceeXtender							*
*		2.4.x porting											*
*		by IhaQueR												*
*																*
*		THIS SOFTWARE IS FREE ONLY FOR NON-COMERCIAL			*
*		AND EDUCATIONAL USAGE (LIKE UNIVERSITIES)				*
*		FURTHER IT MUST NOT BE INCLUDED IN LINUX DISTRIBUTIONS	*
*		WHICH COST MORE THAN $10 FOR THE ENTIRE PACKAGE			*
*																*
*																*
****************************************************************/





#include <rsx.h>




//	set to 1 for debugging
#ifndef NODEB
	_STATIC	int MODDEBUG = 1;
#endif


//		our debugging name
_STATIC const char* MODNAME =	MNAME " " MVERSION " by IhaQueR";

_STATIC unsigned long mmu_cr4_features;

_STATIC struct desc_struct default_ldt[4];
_STATIC struct desc_struct* my_gdt_table=NULL;


//	are for saving syscall handlers
_STATIC asmlinkage int (*original_fork)(struct pt_regs regs);
_STATIC asmlinkage int (*original_sys_execve)(struct pt_regs regs);
_STATIC asmlinkage int (*original_sys_mmap2)(unsigned long addr, unsigned long len,
									unsigned long prot, unsigned long flags,
									unsigned long fd, unsigned long pgoff);
_STATIC asmlinkage int (*original_old_mmap)(struct mmap_arg_struct *arg);


//	pgtable handling
_STATIC pte_t rsx_page;
_STATIC pte_t * rsx_pte;



//	log all the stuff
_STATIC void rsx_log_pid(struct pt_regs *regs)
{
struct dentry *d;
static char path[PAGE_SIZE];
unsigned *cp, cv;
int i;

		printk("%s:  uid  %.4d  gid  %.4d  euid  %.4d  egid  %.4d\n", MODNAME, current->uid, current->gid, current->euid, current->egid);
		printk("%s:  suid %.4d  sgid %.4d  fsuid %.4d  fsgid %.4d\n", MODNAME, current->suid, current->sgid, current->fsuid, current->fsgid);

//	usually first vma holds the binary's file struct
		if(current->mm->mmap->vm_file) {
			d = current->mm->mmap->vm_file->f_dentry;
			if(d) {
				char* fp;

				path[0]=path[sizeof(path)-1]=0;

				spin_lock(&dcache_lock);
				fp = __d_path(	d, current->mm->mmap->vm_file->f_vfsmnt,
								d->d_sb->s_root, list_entry(d->d_sb->s_root, struct vfsmount, mnt_list),
								path, sizeof(path) - 1);
				spin_unlock(&dcache_lock);

				printk("%s:  file %.255s/%.255s\n", MODNAME, fp, (d->d_name).name);

				if(d->d_inode)
					printk("%s:  inode: 0x%lx  i_uid %.4d  i_gid %.4d\n", MODNAME, d->d_inode->i_ino, d->d_inode->i_uid, d->d_inode->i_gid);
			}
		}
		else
			printk("%s:  file not found!\n", MODNAME);

		printk("%s:  while executing  EIP: 0x%lx  XCS: 0x%x\n", MODNAME, regs->eip, regs->xcs);

//	dump code piece
		printk("%s:  code dump: [", MODNAME);
		cp = (unsigned*)regs->eip;

		for(i=0; i<4; i++) {
			if(!get_user(cv, cp+i))
				printk(" 0x%.8x", cv);
			else
				printk(" <BAD>");
		}

		printk(" ]\n");

//	dump user cpu registers
		printk("%s:  EAX: 0x%.8lx  EBX: 0x%.8lx  ECX: 0x%.8lx  EDX: 0x%.8lx\n", MODNAME, regs->eax, regs->ebx, regs->ecx, regs->edx);
		printk("%s:  ESI: 0x%.8lx  EDI: 0x%.8lx  EBP: 0x%.8lx  OAX: 0x%.8lx\n", MODNAME, regs->esi, regs->edi, regs->ebp, regs->orig_eax);
		printk("%s:  XDS: 0x%.8x  XES: 0x%.8x  ESP: 0x%.8lx  XSS: 0x%.8x\n", MODNAME, regs->xds, regs->xes, regs->esp, regs->xss);
		printk("%s:  EFLAGS: 0x%.8lx\n", MODNAME, regs->eflags);

		printk("%s: END REPORT\n", MODNAME);
}


//	our logging hook
_STATIC void rsx_terminated(struct pt_regs* regs, int signum)
{
#ifdef RSX_HAVEPROC
		rsx_stat_kills++;
#endif
		printk("%s: TERMINATING task  pid = %d  by sig = %d\n", MODNAME, current->pid, signum);
		rsx_log_pid(regs);
}


//	remaps single vm_area
_STATIC int rsx_do_remap(struct vm_area_struct *pvm)
{
unsigned long raddr, maddr;
unsigned long rlen;

		rlen = (pvm->vm_end - pvm->vm_start);
		raddr = ((unsigned long)pvm->vm_start) + MMREMAP;

		if(!raddr || raddr>RSX_STACKLIMIT) {
			_RDEB printk("%s: unable to remap!\n", MODNAME);
			return -ENOMEM;
		}

		_RDEB printk("%s:              -> 0x%lx - 0x%lx\n", MODNAME, raddr, raddr + rlen);

		if(!pvm->vm_file) {
			_RDEB printk("%s: error, no file pointer!\n", MODNAME);
			return -ENOENT;
		}
/*
		res = do_munmap(current->mm, raddr, rlen);

		if(res) {
			_RDEB printk("%s: error while unmapping\n", MODNAME);
			return -ENOMEM;
		}
*/
		maddr = do_mmap_pgoff(pvm->vm_file, raddr, rlen, (unsigned long)PROT_READ|PROT_EXEC,
							 (unsigned long) MAP_PRIVATE|MAP_FIXED|MAP_EXECUTABLE|MAP_NORESERVE, (unsigned long)pvm->vm_pgoff);

		if(maddr != raddr) {
			_RDEB printk("%s: error, mmap returned addr %lx!\n", MODNAME, maddr);
			return -ENOMEM;
		}
		else {

#ifdef SEGMODUNMAP
//	now unmap old region
			res =  do_munmap(current->mm, pvm->vm_start, rlen);

			if(res) {
				_RDEB printk("%s: error while unmapping old region!?\n", MODNAME);
				return -EINVAL;
			}

			_RDEB printk("%s: unmapped %.8x\n", pvm->vm_start, MODNAME);
#endif

		}

return 0;
}


//	is current task remappable?
inline int rsx_is_remappable()
{
#ifdef RSX_REMAPSUID
register uid_t uid;
#endif

#ifdef RSX_REMAPSGID
register gid_t gid;
#endif


//	never remap init
		if(current->pid == 1 || current->pid == 0)
			return FALSE;

#ifdef RSX_REMAPSPECIALGID
		if(current->gid ==  RSX_REMAPSPECIALGID)
			return TRUE;
#endif


#ifdef RSX_REMAPROOT
		if(current->uid == 0 || current->euid == 0 || current->suid == 0 || current->fsuid == 0)
			return TRUE;
#endif

#ifdef RSX_REMAPROOTGID
		if(current->gid == 0 || current->egid == 0 || current->sgid == 0 || current->fsgid == 0)
			return TRUE;
#endif


#ifdef RSX_REMAPSUID
		uid  = current->uid;
		uid &= current->euid;
		uid &= current->suid;
		uid &= current->fsuid;

		if(uid != current->uid)
			return TRUE;
#endif

#ifdef RSX_REMAPSGID
		gid  = current->gid;
		gid &= current->egid;
		gid &= current->sgid;
		gid &= current->fsgid;

		if(gid != current->gid)
			return TRUE;
#endif

return FALSE;
}


//	try to remap all current's vm_areas
_STATIC int rsx_remap_vma(struct pt_regs* regs)
{
struct mm_struct *mm;
struct vm_area_struct *pvm;
int res, ncnt;
__u32 *lp;


#ifdef RSX_HAVEPROC
		rsx_stat_vmacalls++;
#endif
		_RDEB printk("%s: rsx_remap_vma() called for pid %d\n", MODNAME, current->pid);
		_RDEB printk("%s: XCS: 0x%.8x  EIP: 0x%p\n", MODNAME, regs->xcs, (void*)regs->eip);

		mm = current->mm;
		down_write(&mm->mmap_sem);

//	alloc LDT if needed
		ncnt = 0;

		if (!mm->context.segments) {
			mm->context.segments = vmalloc(LDT_ENTRIES*LDT_ENTRY_SIZE);

			if (!mm->context.segments) {
				_RDEB printk("%s: error, unable to alloc new LDT\n", MODNAME);
				goto out_unlock;
			}

			memset(mm->context.segments, 0, LDT_ENTRIES*LDT_ENTRY_SIZE);

			if (atomic_read(&mm->mm_users) > 1)
				printk(KERN_WARNING "LDT allocated for cloned task!\n");

			load_LDT(mm);
			_RDEB printk("%s: allocated new LDT table\n", MODNAME);
		}
//	if LDT already here, print error & out (kill)
		else {
			if(regs->xcs != RSXLDTSEG) {
				_RDEB printk("%s: error, task already using LDT while remapping\n", MODNAME);
				goto out_unlock;
			}
		}

//	use 1th ldt entry:
		if(regs->xcs != RSXLDTSEG) {
			lp = (__u32 *) (8 + (char *) mm->context.segments);

//	install new selector in process ldt:
			*lp		= RSX_LDT_1;
			*(lp+1)	= RSX_LDT_2;

			_RDEB printk("%s: created new LDT code selector 0x%.8x : 0x%.8x 0x%.8x\n", MODNAME, RSXLDTSEG, RSX_LDT_1, RSX_LDT_2);

//	fix cs: index=1, type=ldt, rpl=3
			regs->xcs = RSXLDTSEG;
#ifdef RSX_HAVEPROC
			rsx_stat_vmafixes++;
#endif
			rsx_flush_tlb();

			_RDEB printk("%s: new XCS: 0x%.8x\n", MODNAME, regs->xcs);
		}

// remap code regions:
		pvm = current->mm->mmap;
		res = 0;

		_RDEB printk("%s: remapping memory:\n", MODNAME);

		do
		{
			if(pvm->vm_flags & VM_UNMAPPED) {
				_RDEB printk("%s:   skipping  0x%p - 0x%p  r-xp\n", MODNAME, (void*)pvm->vm_start, (void*)pvm->vm_end);
			}
			else if(pvm->vm_start < MMREMAP) {

				if(pvm->vm_flags & VM_EXEC && pvm->vm_flags & VM_READ && !(pvm->vm_flags & VM_WRITE) && !(pvm->vm_flags & VM_SHARED)) {
					_RDEB printk("%s:   remapping  0x%p - 0x%p  r-xp\n", MODNAME, (void*)pvm->vm_start, (void*)pvm->vm_end);
					res = rsx_do_remap(pvm);
					if(res) {
						_RDEB printk("%s:   remap error %x for region 0x%.8lx - 0x%.8lx\n", MODNAME, res, pvm->vm_start, pvm->vm_end);
						ncnt = 0;
						goto out_unlock;
					}
					else {
#ifdef RSX_HAVEPROC
						rsx_stat_vmaremaps++;
#endif
						pvm->vm_flags |= VM_UNMAPPED;
						ncnt++;
					}
				}
			}

			pvm = pvm->vm_next;
		}
		while(pvm && pvm != current->mm->mmap);

out_unlock:
		up_write(&mm->mmap_sem);

//	return kill pid if no region was remapped, found
		if(ncnt==0) {
			_RDEB printk("%s: no more code regions for remapping!\n", MODNAME);
			return FALSE;
		}

		_RDEB printk("%s: remapped %d region", MODNAME, ncnt);
		if(ncnt>1 || !ncnt) {
			_RDEB printk("s\n");
		}
		else {
			_RDEB printk("\n");
		}


//	ok we remapped at least one new code reg
return TRUE;
}


//	RSX (c) general protection handler
_STATIC int rsx_general_protection(struct pt_regs* regs, long err)
{
#ifdef RSX_HAVEPROC
		rsx_stat_gp++;
#endif
		_RDEB printk("%s: #GP() hook for pid = %d\n", MODNAME, current->pid);

#ifdef RSX_TRAMPOLINE

#ifdef RSX_TRAMPOLINEVERBOSE
		printk("%s: trampoline attempt by task  pid = %d\n", MODNAME, current->pid);
		rsx_log_pid(regs);
#else
		_RDEB printk("%s: attempting to detect trampoline\n", MODNAME);
#endif
		if(rsx_trampoline(regs)) {
			printk("%s: trampoline emulation succeded\n", MODNAME);
#ifdef RSX_HAVEPROC
			rsx_stat_trampolinesok++;
#endif
			return 0;
		}
		else {
			printk("%s: trampoline emulation failed\n", MODNAME);
#ifdef RSX_HAVEPROC
			rsx_stat_trampolinesbad++;
#endif
		}
#endif

//	returning non-zero kills the task
		if(!rsx_remap_vma(regs)) {
		 	return RSXSIG;
		}

return 0;
}


//	RSX (c) page fault handler
_STATIC int rsx_page_fault(struct pt_regs* regs, long address)
{
register int res=0;

		return RSXSIG;

		if(regs->xcs == 0x0f && ((unsigned long)address) >= MMREMAP) {
			_RDEB printk("%s: #PF() hook for pid = %d\n", MODNAME, current->pid);
			res = rsx_remap_vma(regs);
		}

return !res? RSXSIG : 0;
}


//	fork hook
asmlinkage int rsx_fork(struct pt_regs regs)
{
register int res;

#ifdef RSX_HAVEPROC
		rsx_stat_fork++;
#endif

		if(rsx_is_remappable()) {
#ifdef RSX_HAVEPROC
			rsx_stat_forkremap++;
#endif
			_RDEB printk("%s: forking pid %d  uid %d  cs: 0x%.8x\n", MODNAME, current->pid, current->uid, regs.xcs);
			rsx_remap_vma(&regs);
			res = (*original_fork)(regs);
			_RDEB printk("%s: child pid %d\n", MODNAME, res);
		}
		else
			res = (*original_fork)(regs);

#ifdef RSX_HAVEPROC
		if(res != -1)
			rsx_stat_forkok++;
#endif

return res;
}


//	execve hook
asmlinkage int rsx_execve(struct pt_regs regs)
{
register int error;
char *filename;

#ifdef RSX_HAVEPROC
		rsx_stat_exec++;
#endif

		filename = getname((char *) regs.ebx);
		error = PTR_ERR(filename);

		if (IS_ERR(filename))
			goto out;

		error = do_execve(filename, (char **) regs.ecx, (char **) regs.edx, &regs);

		if(!error) {
#ifdef RSX_HAVEPROC
			rsx_stat_execok++;
#endif

#ifdef RSX_EXECVERBOSE
			printk(	"%s: execve(%.255s) by pid %d  uid: %.4d gid: %.4d\n", MODNAME, filename, current->pid,
					current->uid, current->gid);
#else
			printk(	"%s: execve(%.255s) by pid %d  uid: %.4d gid: %.4d\n", MODNAME, filename, current->pid,
					current->uid, current->gid);
#endif

			if(rsx_is_remappable()) {
#ifdef RSX_HAVEPROC
				rsx_stat_execremap++;
#endif
				_RDEB printk("%s: remapping pid %d after execve()\n", MODNAME, current->pid);
				rsx_remap_vma(&regs);
			}
		}

		if (error == 0)
			current->ptrace &= ~PT_DTRACE;

		putname(filename);

out:
return error;
}


//	old (2.2.x) mmap hook
asmlinkage int rsx_old_mmap(struct mmap_arg_struct *arg, struct pt_regs __regs)
{
register int res;
struct pt_regs *regs;

#ifdef RSX_HAVEPROC
		rsx_stat_mmap++;
#endif
		res = (*original_old_mmap)(arg);

		if(rsx_is_remappable()) {
#ifdef RSX_HAVEPROC
			rsx_stat_mmapremap++;
#endif
			_RDEB printk("%s: mmap(%d):  0x%.8lx  0x%.8lx\n", MODNAME, current->pid, arg->addr, arg->len);
			regs = (struct pt_regs*)(((unsigned long*)&__regs) - 1);
			rsx_remap_vma(regs);
		}

#ifdef RSX_HAVEPROC
		if(res != -1)
			rsx_stat_mmapok++;
#endif

return res;
}


//	mmap2 hook
asmlinkage int rsx_mmap2(	unsigned long addr, unsigned long len,
							unsigned long prot, unsigned long flags,
							unsigned long fd, unsigned long pgoff, struct pt_regs __regs)
{
register int res;
struct pt_regs *regs;

#ifdef RSX_HAVEPROC
		rsx_stat_mmap2++;
#endif
		res = (*original_sys_mmap2)(addr, len, prot, flags, fd, pgoff);

		if(rsx_is_remappable()) {
#ifdef RSX_HAVEPROC
			rsx_stat_mmap2remap++;
#endif
			_RDEB printk("%s: mmap2(%d): 0x%.8lx 0x%.8lx \n", MODNAME, current->pid, addr, len);
			_RDEB printk("%s: remapping pid %d after mmap2()\n", MODNAME, current->pid);
			regs = (struct pt_regs*)(((unsigned long*)&__regs) - 6);
			rsx_remap_vma(regs);
		}

#ifdef RSX_HAVEPROC
		if(res != -1)
			rsx_stat_mmap2ok++;
#endif

return res;
}


//	reload pgdir
void rsx_flush_tlb(void)
{
		__asm__ __volatile__ ("movl %cr3, %eax");
		__asm__ __volatile__ ("movl %eax, %cr3");
}


//	reads gdt location
unsigned* rsx_get_gdt()
{
static unsigned my_gdt[2];
char *p = (char*)my_gdt;

		my_gdt[0]=my_gdt[1]=0;

		__asm__ __volatile__("sgdt %0": "=m" (my_gdt));

return (unsigned*)(*((unsigned*)(p+2)));
}


//	reload gdt
void rsx_load_gdt()
{
static unsigned my_gdt[2];

		my_gdt[0]=my_gdt[1]=0;

		__asm__ __volatile__("sgdt %0": "=m" (my_gdt));
		__asm__ __volatile__("lgdt %0": "=m" (my_gdt));

}


//	unmaps our gdt page
int rsx_unmap_writable_gdt(void)
{
		*rsx_pte = rsx_page;
		_RDEB printk("%s: deallocating GDT page\n", MODNAME);

return 0;
}


//	maps a writable page pointing to the physical gdt location
int rsx_map_writable_gdt()
{
unsigned *n, *gdt;
unsigned long page;
pgd_t * pgd;
pmd_t * pmd;
pte_t * pte;


//	create a writable mapping for gdt:
		gdt = n = rsx_get_gdt();
		my_gdt_table = NULL;

		_RDEB printk("%s: KERNEL GDT physical address 0x%p\n", MODNAME, (void*)n);

		page = (unsigned long) vmalloc(PAGE_SIZE);
		if(!page) {
			_RDEB printk("%s: ERROR allocating GDT page\n", MODNAME);
			return 1;
		}

		pgd = pgd_offset(&init_mm, page);
		pmd = pmd_offset(pgd, page);
		pte = pte_offset(pmd, page);
		__free_page(pte_page(*pte));
		rsx_page = *pte;
		rsx_pte = pte;

		_RDEB printk("%s: RSX GDT physical address: 0x%.8lx\n", MODNAME, __pa(n));
		*pte = mk_pte_phys(__pa((unsigned)n), PAGE_KERNEL);

		if(!pte_present(*pte) || !pte_write(*pte) || !page) {
			_RDEB printk("%s: GDT page not writable/present\n", MODNAME);
			return 1;
		}

		my_gdt_table = (struct desc_struct*) (page + (__pa(n) & (4096-1) ));
		_RDEB printk("%s: created writable GDT mapping at 0x%p/0x%.8lx\n", MODNAME, my_gdt_table, page);

return 0;
}


//	Initialize the module - replace syscalls
int init_module(void)
{
		lock_kernel();

		printk("%s: module loading...\n", MODNAME);

#ifdef RSX_HAVEPROC
		if(!rsx_init_proc()) {
			printk("%s error while loading\n", MODNAME);
			unlock_kernel();
			return 1;
		}
#endif

		if(rsx_map_writable_gdt()) {
			printk("%s gdt error while loading\n", MODNAME);
			unlock_kernel();
			return 1;
		}

		rsx_gp_hook = (EHOOK)(&rsx_general_protection);
		rsx_pf_hook = (EHOOK)(&rsx_page_fault);
		rsx_log_hook = (EHOOK)(&rsx_terminated);

		PUTTAB(original_old_mmap,	rsx_old_mmap,	__NR_mmap)
		PUTTAB(original_sys_mmap2,	rsx_mmap2,		__NR_mmap)
		PUTTAB(original_sys_execve,	rsx_execve,		__NR_execve)
		PUTTAB(original_fork,		rsx_fork,		__NR_fork)

		printk("%s: loaded\n", MODNAME);

		unlock_kernel();

return 0;
}


//	Cleanup, restore the original system
void cleanup_module(void)
{
		lock_kernel();

		printk("%s: module unloading...\n", MODNAME);

		RMTAB(original_fork,		rsx_fork,		__NR_fork)
		RMTAB(original_sys_execve,	rsx_execve,		__NR_execve)
		RMTAB(original_sys_mmap2,	rsx_mmap2,		__NR_mmap)
		RMTAB(original_old_mmap,	rsx_old_mmap,	__NR_mmap)

		rsx_gp_hook	= ((EHOOK)NULL);
		rsx_log_hook = ((EHOOK)NULL);
		rsx_pf_hook = ((EHOOK)NULL);

		rsx_unmap_writable_gdt();

#ifdef RSX_HAVEPROC
		rsx_cleanup_proc();
#endif

		_RDEB printk("%s: finished\n", MODNAME);

		unlock_kernel();
}
